from django.test.testcases import TestCase
from django.contrib.auth.models import Group, User, Permission
from django.contrib.contenttypes.models import ContentType
from djinn_auth.models import Role
from djinn_auth.utils import assign_global_role, unassign_global_role, \
    assign_local_role, unassign_local_role
from djinn_auth.authbackend import AuthBackend


class AuthBackendTest(TestCase):

    """ Test the auth backend for role base permissions. """

    def setUp(self):

        self.backend = AuthBackend()
        self.owner, created = Role.objects.get_or_create(name="owner")

        ctype = ContentType.objects.get_for_model(Group)

        do_something, created = Permission.objects.get_or_create(
            codename="do_something",
            content_type=ctype, defaults={'name': 'Can do things'})

        self.perm = do_something
        self.owner.add_permission(do_something)

    def test_has_perm_by_global_role(self):

        """Test global permissions when a global role has been granted to the
        user
        """

        tjibbe = User.objects.create(username="Tjibbe")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

        assign_global_role(tjibbe, self.owner)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something"))

        unassign_global_role(tjibbe, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

    def test_has_perm_by_permission_on_user(self):

        """Test global permissions when a specific permission has been granted
        to the user
        """

        tjibbe = User.objects.create(username="Tjibbe")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbe.user_permissions.add(self.perm)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbe.user_permissions.remove(self.perm)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

    def test_has_perm_by_global_role_on_group(self):

        """Test global permissions when the permission is part of a global
        group role

        """

        tjibbe = User.objects.create(username="Tjibbe")
        tjibbes = Group.objects.create(name="Tjibbes")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

        assign_global_role(tjibbes, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbe.groups.add(tjibbes)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbe.groups.remove(tjibbes)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

    def test_has_perm_by_global_permission_on_group(self):

        """ Test global permissions when the permission has been set
        onto a group that the user is a member of """

        tjibbe = User.objects.create(username="Tjibbe")
        tjibbes = Group.objects.create(name="Tjibbes")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbes.permissions.add(self.perm)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbe.groups.add(tjibbes)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something"))

        tjibbe.groups.remove(tjibbes)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something"))

    def test_has_perm_on_object(self):

        """ Check whether the user has the permission on the object, when
        a global role has been assigned to this user."""

        tjibbe = User.objects.create(username="Tjibbe")

        content = Group.objects.create(name="Tjibbes")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        assign_global_role(tjibbe, self.owner)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something",
                                              obj=content))

        unassign_global_role(tjibbe, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

    def test_has_perm_on_object_with_userpermission(self):

        """ Test whether the permission is there when the user has been given
        a specific permission."""

        tjibbe = User.objects.create(username="Tjibbe")

        content = Group.objects.create(name="Tjibbes")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        tjibbe.user_permissions.add(self.perm)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something",
                                              obj=content))

        tjibbe.user_permissions.remove(self.perm)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

    def test_has_perm_on_object_by_local_role(self):

        tjibbe = User.objects.create(username="Tjibbe")

        content = Group.objects.create(name="Tjibbes")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        assign_local_role(tjibbe, content, self.owner)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something",
                                              obj=content))

        unassign_local_role(tjibbe, content, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

    def test_has_perm_on_object_by_local_role_on_group(self):

        tjibbe = User.objects.create(username="Tjibbe")
        tjibbes = Group.objects.create(name="Tjibbes")

        content = Group.objects.create(name="Content")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        assign_local_role(tjibbes, content, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        tjibbe.groups.add(tjibbes)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something",
                                              obj=content))

        tjibbe.groups.remove(tjibbes)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

    def test_local_roles_without_acquire_global(self):

        """Test whether the backend actually honors the 'acquire_global_roles'
        setting.  """

        tjibbe = User.objects.create(username="Tjibbe")
        content = Group.objects.create(name="Content")

        assign_global_role(tjibbe, self.owner)

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something",
                                              obj=content))

        setattr(content, "acquire_global_roles", False)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

    def test_has_perm_with_acquisition(self):

        content = Group.objects.create(name="Tjibbes")
        parent = Group.objects.create(name="Parent")

        tjibbe = User.objects.create(username="Tjibbe")

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        assign_local_role(tjibbe, parent, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        assign_local_role(tjibbe, parent, self.owner)

        self.assertFalse(self.backend.has_perm(tjibbe, "app.do_something",
                                               obj=content))

        content.acquire_from = [parent]

        self.assertTrue(self.backend.has_perm(tjibbe, "app.do_something",
                                              obj=content))



import logging
from django.contrib.auth.models import Permission
from djinn_auth.utils import get_user_global_roles, get_user_local_roles


LOGGER = logging.getLogger("djinn_auth")


class AuthBackend(object):

    # We support object based permissions
    #
    supports_object_permissions = True
    supports_anonymous_user = False

    def authenticate(self, username=None, password=None):

        return None

    def get_user(self, user_id):

        return None

    def has_perm(self, user, permission, obj=None):

        """ Find out whether the user has a given permission. This is
        a potentially 'heavy' operation, given the role based auth
        that we have.

        A user can have a permission in general (obj=None) if:
          1. the user has a global role that holds the permission
          2. the user is part of a group that has a global role that holds the
             permission

        When an object is provided, the user can FURTHERMORE have a permission
        if:
          3. the user has a local role on the object
          4. the user is part of a group that has a local role on the object

        Some heuristics are in place to try and determine this in a
        smart way. We first check on the 'user' role. If this holds
        the permission, return. If not, check ownership...

        """

        if user.is_anonymous():
            return False

        return self._check_all_permissions(user, permission, obj=obj)

    def _check_all_permissions(self, user, perm, obj=None):

        """ Go find the actual permission, then loop over it's roles, users
        and groups. Check if the user has the global role, the role
        locally if obj is provided, or is in a group with that
        permission.
        """

        perm_app, perm_name = perm.split(".")

        _perm = Permission.objects.get(codename=perm_name)

        # Check whether the user is in the permission's user set
        #
        _filter = {user.USERNAME_FIELD: user.get_username()}

        if _perm.user_set.filter(**_filter).exists():
            return True

        user_group_ids = user.groups.all().values_list('id', flat=True)

        # Check whether the user and the permission share any groups
        #
        if _perm.group_set.filter(pk__in=user_group_ids):
            return True

        # Now check on the roles. Start with global roles
        #
        perm_role_ids = _perm.role_set.all().values_list('id', flat=True)

        if not obj or getattr(obj, "acquire_global_roles", True):

            if get_user_global_roles(user).filter(
                    role__id__in=perm_role_ids).exists():
                return True

        # Now go for local roles if need be
        #
        if obj:

            if get_user_local_roles(user, obj).filter(
                    role__id__in=perm_role_ids
            ).exists():
                return True

        # Finally, we may have an acquire list, so as to be able to
        # 'inherit' roles from another object, or objects.
        #
        for acq_obj in getattr(obj, "acquire_from", []):
            if get_user_local_roles(user, acq_obj).filter(
                    role__id__in=perm_role_ids
            ).exists():
                return True

        return False



from django.test.testcases import TestCase
from django.contrib.auth.models import User
from djinn_auth.models import Role
from djinn_auth.utils import get_global_roles, has_global_role, \
    assign_global_role, unassign_global_role


class GlobalRoleTest(TestCase):

    def setUp(self):

        self.role = Role.objects.create(name="somerole")

    def test_assign_global_role(self):

        tjibbe = User.objects.create(username="Tjibbe")
        gurbe = User.objects.create(username="Gurbe")

        self.assertFalse(has_global_role(tjibbe, self.role))
        self.assertFalse(has_global_role(gurbe, self.role))

        assign_global_role(tjibbe, self.role)

        self.assertTrue(has_global_role(tjibbe, self.role))
        self.assertFalse(has_global_role(gurbe, self.role))

    def test_has_global_role(self):

        tjibbe = User.objects.create(username="Tjibbe")

        self.assertFalse(has_global_role(tjibbe, self.role))

        assign_global_role(tjibbe, self.role)

        self.assertTrue(has_global_role(tjibbe, self.role))
        self.assertTrue(has_global_role(tjibbe, "somerole"))

    def test_unassign_global_role(self):

        tjibbe = User.objects.create(username="Tjibbe")

        self.assertFalse(has_global_role(tjibbe, self.role))

        assign_global_role(tjibbe, self.role)

        self.assertTrue(has_global_role(tjibbe, self.role))

        unassign_global_role(tjibbe, self.role)

        self.assertFalse(has_global_role(tjibbe, self.role))

    def test_get_global_roles(self):

        tjibbe = User.objects.create(username="Tjibbe")

        self.assertEquals(0, len(get_global_roles(tjibbe)))

        assign_global_role(tjibbe, self.role)

        self.assertEquals(1, len(get_global_roles(tjibbe)))



from django.template import Library
from django.template.base import NodeList
from django.template.defaulttags import IfNode, TemplateLiteral


register = Library()


class HasPermissionNode(IfNode):

    def __init__(self, ctx, user, perm, nodelist_true, nodelist_false=None):
        self.nodelist_true, self.nodelist_false = nodelist_true, nodelist_false
        self.ctx = ctx
        self.usr = user
        self.perm = perm

    def render(self, context):

        ctx = self.ctx.eval(context)
        usr = self.usr.eval(context)

        if hasattr(self.perm, "eval"):
            perm = self.perm.eval(context)
        else:
            perm = self.perm

        # Normalize permission if need be...
        #
        if ctx and perm in ["delete", "change", "add", "view"]:
            perm = "%s.%s_%s" % (ctx._meta.app_label, perm,
                                 ctx._meta.object_name.lower())

        if usr.has_perm(perm, obj=ctx):
            return self.nodelist_true.render(context)
        else:
            return self.nodelist_false.render(context)


def if_has_perm(parser, token):

    """ Usage: if_has_perm <context> <user object> '<permission codename>'"""

    bits = token.split_contents()[1:]
    ctx = TemplateLiteral(parser.compile_filter(bits[0]), bits[0])
    user = TemplateLiteral(parser.compile_filter(bits[1]), bits[1])
    perm = TemplateLiteral(parser.compile_filter(bits[2]), bits[2])

    nodelist_true = parser.parse(('else', 'endif_has_perm'))
    token = parser.next_token()

    if token.contents == 'else':
        nodelist_false = parser.parse(('endif_has_perm',))
        parser.delete_first_token()
    else:
        nodelist_false = NodeList()
    return HasPermissionNode(ctx, user, perm, nodelist_true, nodelist_false)

register.tag("if_has_perm", if_has_perm)



from djinn_auth.models.base import RoleAssignment


class GlobalRole(RoleAssignment):

    """Provide a means of giving a user or group a specific role that
    is global, so it will add it's permissions on ALL calls on for
    instance 'has_perm'.

    """

    class Meta:

        app_label = "djinn_auth"



from django.db import models
from django.conf import settings
from django.contrib.auth.models import Permission
from polymorphic import PolymorphicModel


class Group(PolymorphicModel):

    """Abstract group base. The polymorphism enables you to extend this
    class in several ways, but find all kinds with the Group.objects
    manager

    """

    name = models.CharField(max_length=200)
    description = models.TextField(null=True, blank=True)
    users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        blank=True)
    permissions = models.ManyToManyField(
        Permission,
        related_name='groups',
        blank=True)

    def __unicode__(self):

        return self.name

    class Meta:
        ordering = ["name"]
        app_label = "djinn_auth"
